/*
     This file is part of GNUnet.
     (C) 2005 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/main.c
 * @brief Main function of gnunet-gtk
 * @author Christian Grothoff
 */

#include "config.h"
#include "gettext.h"

#include <GNUnet/gnunet_util.h>

#include "main.h"
#include "fs.h"
#include "daemon.h"
#include "helper.h"

/**
 * Prints the usage information for this command if the user errs.
 */
static void printhelp() {
  static Help help[] = {
    HELP_CONFIG,
    HELP_HELP,
    HELP_HOSTNAME,
    HELP_LOGLEVEL,
    HELP_VERSION,
    HELP_END,
  };
  formatHelp("gnunet-gtk [OPTIONS]",
	     _("GNUnet GTK user interface."),
	     help);
}

/**
 * Parse the options, set the timeout.
 * @param argc the number of options
 * @param argv the option list (including keywords)
 * @return SYSERR if we should exit, OK otherwise
 */
static int parseOptions(int argc,
			char ** argv) {
  int c;

  while (1) {
    int option_index = 0;
    static struct GNoption long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { 0,0,0,0 }
    };
    c = GNgetopt_long(argc,
		      argv,
		      "c:dhH:L:v",
		      long_options,
		      &option_index);
    if (c == -1)
      break;  /* No more flags to process */
    if (YES == parseDefaultOptions(c, GNoptarg))
      continue;
    switch(c) {
    case 'h':
      printhelp();
      return SYSERR;
    case 'v':
      printf("GNUnet v%s, gnunet-gkt v%s\n",
	     VERSION,
	     AFS_VERSION);
      return SYSERR;
    default:
      LOG(LOG_FAILURE,
	  _("Use --help to get a list of options.\n"));
      return SYSERR;
    } /* end of parsing commandline */
  } /* while (1) */
  setConfigurationStringList(&argv[GNoptind],
			     argc-GNoptind);
  return OK;
}

GladeXML * mainXML;
char *gladeFile = NULL;

int main(int argc, 
	 char *argv[]) {
  
  g_thread_init(NULL);
  gtk_init(&argc, &argv);
  if (SYSERR == initUtil(argc, 
			 argv, 
			 &parseOptions))
    return 0;
  startCron();
  gtkInitSaveCalls();
  /* load the interface */
#ifdef MINGW
	gladeFile = MALLOC(_MAX_PATH + 1);
	plibc_conv_to_win_path(PACKAGE_DATA_DIR"/gnunet-gtk.glade", gladeFile);
#else
	gladeFile = STRDUP(PACKAGE_DATA_DIR"/gnunet-gtk.glade");
#endif

  mainXML = glade_xml_new(gladeFile,
			  "mainWindow", 
			  NULL);
			  
  /* connect the signals in the interface */
  setCustomLogProc(&addLogEntry);
  glade_xml_signal_autoconnect(mainXML);
  gtk_fs_init();
  addCronJob(&cronCheckDaemon, 
	     0,
	     15 * cronSECONDS,
	     NULL);
  /* start the event loop */
  gdk_threads_enter();  
  gtk_main();
  gdk_threads_leave(); 
  setCustomLogProc(NULL);
  stopCron();
  delCronJob(&cronCheckDaemon, 
	     15 * cronSECONDS,
	     NULL);
  gtk_fs_done();
  gtkDoneSaveCalls();
  UNREF(mainXML);
  doneUtil();
  FREE(gladeFile);
  
  return 0;
}

/* ************* end of main.c ************ */
